﻿#region Usings
using System;
using System.Diagnostics;
using System.Linq;
using System.Net;
using System.Threading;
using Microsoft.WindowsAzure.Diagnostics;
using Microsoft.WindowsAzure.ServiceRuntime;
using Microsoft.WindowsAzure;
using Microsoft.WindowsAzure.StorageClient;
using Core.Logic;
using Core.Types;
using System.IO;
using System.Xml.Serialization;
#endregion

namespace BankWorkerRole2
{
  public class WorkerRole : RoleEntryPoint
  {
    private CloudQueue _accountTransactionQueue=null;

    #region Configuration stuff
    
    public override void Run()
    {
      // This is a sample worker implementation. Replace with your logic.
      Trace.WriteLine("BankWorkerRole2 entry point called", "Information");

      while (true)
      {
        CloudQueueMessage msg = _accountTransactionQueue.GetMessage(new TimeSpan(0,0,20));

        if (msg != null)
        {
          ProcessMessage(_accountTransactionQueue, msg);
          Trace.WriteLine(String.Format("Deleted message, message id={0}",msg.Id),"Information");
        }
        else
        {
          // poll again later
          Thread.Sleep(1000);
        }
      }
    }

    public override bool OnStart()
    {
      ServicePointManager.DefaultConnectionLimit = 12;
      // Set the maximum number of concurrent connections 
      ServicePointManager.DefaultConnectionLimit = 12;

      // Get the factory configuration so that it can be edited
      DiagnosticMonitorConfiguration config = DiagnosticMonitor.GetDefaultInitialConfiguration();

      // Set scheduled transfer interval for infrastructure logs to 1 minute
      config.DiagnosticInfrastructureLogs.ScheduledTransferPeriod = System.TimeSpan.FromMinutes(1);

      // Specify a logging level to filter records to transfer
      config.DiagnosticInfrastructureLogs.ScheduledTransferLogLevelFilter = LogLevel.Error;

      // Set scheduled transfer interval for user's Windows Azure Logs to 1 minute
      config.Logs.ScheduledTransferPeriod = System.TimeSpan.FromMinutes(1);

      DiagnosticMonitor.Start("DiagnosticsConnectionString", config);

      // For information on handling configuration changes
      // see the MSDN topic at http://go.microsoft.com/fwlink/?LinkId=166357.
      RoleEnvironment.Changing += RoleEnvironmentChanging;

      CloudStorageAccount.SetConfigurationSettingPublisher((configName, configSetter) =>
      {
        configSetter(RoleEnvironment.GetConfigurationSettingValue(configName));
        RoleEnvironment.Changed += (sender, arg) =>
        {
          if (arg.Changes.OfType<RoleEnvironmentConfigurationSettingChange>()
              .Any((change) => (change.ConfigurationSettingName == configName)))
          {
            if (!configSetter(RoleEnvironment.GetConfigurationSettingValue(configName)))
            {
              RoleEnvironment.RequestRecycle();
            }
          }
        };
      });

      InitQueueStorage();

      return base.OnStart();
    }

    private void InitQueueStorage()
    {
      var storageAccount = CloudStorageAccount.FromConfigurationSetting("DataConnectionString");
      // initialize queue storage 
      CloudQueueClient queueStorage = storageAccount.CreateCloudQueueClient();
      _accountTransactionQueue = queueStorage.GetQueueReference(Core.Helpers.Queues.AccountTransactionsQueue);

      Trace.TraceInformation("Creating account transaction logg queue...");

      bool storageInitialized = false;
      while (!storageInitialized)
      {
        try
        {
          // create the message queue
          _accountTransactionQueue.CreateIfNotExist();
          storageInitialized = true;
        }
        catch (StorageClientException e)
        {
          if (e.ErrorCode == StorageErrorCode.TransportError)
          {
            Trace.TraceError("Storage services initialization failure. "
              + "Check your storage account configuration settings. If running locally, "
              + "ensure that the Development Storage service is running. Message: '{0}'", e.Message);
            System.Threading.Thread.Sleep(5000);
          }
          else
          {
            throw;
          }
        }
      }
    }

    private void RoleEnvironmentChanging(object sender, RoleEnvironmentChangingEventArgs e)
    {
      // If a configuration setting is changing
      if (e.Changes.Any(change => change is RoleEnvironmentConfigurationSettingChange))
      {
        // Set e.Cancel to true to restart this role instance
        e.Cancel = true;
      }
    }

    #endregion

    private static void ProcessMessage(CloudQueue queue,CloudQueueMessage msg)
    {
      if (!msg.AsString[0].Equals('<'))
      {
        HandleStringMessage(queue, msg);
      }
      else
      {
        HandleXmlMessage(queue, msg);
      }
    }

    private static void HandleXmlMessage(CloudQueue queue, CloudQueueMessage msg)
    {
      // do stuff
      Trace.WriteLine(String.Format("Processing message, id={0} ",msg.Id), "Information");
      using (MemoryStream m = new MemoryStream(msg.AsBytes))
      {
        if (m != null)
        {
          XmlSerializer xs = new XmlSerializer(typeof(Core.TableStorage.UserAccountTransaction));
          try
          {
            var transactionMessage = xs.Deserialize(m) as Core.TableStorage.UserAccountTransaction;
            if (transactionMessage != null)
            {
              UserAccountStorageAction ds = new UserAccountStorageAction();
              var transresult = ds.InsertOrRedrawMoneyWithTableStorage(transactionMessage);
              if (transresult.TransactionStatus == TransactionStatus.Succeded)
              {
                transactionMessage.Message = String.Format("Proccessed {0}, PartKey={1}, Msg={2}", transactionMessage.Amount, transactionMessage.PartitionKey, transresult.Message);

                //userid|accountid|transactionid|amount
                Core.Queues.QueueManager.AccountTransactionsLoggQueue.AddMessage(new CloudQueueMessage(String.Format("{0}|{1}|{2}", transactionMessage.PartitionKey, transactionMessage.RowKey, transactionMessage.Amount)));
                
                Trace.WriteLine(msg.AsString, "Information");
                
                queue.DeleteMessage(msg);
              }
              else if (transresult.TransactionStatus == TransactionStatus.Completed)
              {
                Trace.WriteLine(String.Format("Amount:{0}, Message= {1}, Msg= {2}", transactionMessage.Amount, transactionMessage.Message, transresult.Message), "Warning");
                
                queue.DeleteMessage(msg);
              }
            }
          }
          catch (Exception ex)
          {
            Trace.WriteLine(ex.Message, "Error");
            throw ex;
          }
        }
      }
    }

    private static void HandleStringMessage(CloudQueue queue, CloudQueueMessage msg)
    {
      var str = msg.AsString.Split('|');
      if (str.Length == 4)
      {
        //userid|accountid|transactionid|amount
        UserAccountStorageAction ds = new UserAccountStorageAction();
        var t = new Core.TableStorage.UserAccountTransaction(str[0], str[1], str[2]) { Amount = double.Parse(str[3]) };
        var transresult = ds.InsertOrRedrawMoneyWithTableStorage(t);

        if (transresult.TransactionStatus == TransactionStatus.Succeded)
        {
          //send message to another queue to update db
          //userid|accountid|transactionid|amount
          t.Message = String.Format("Proccessed {0}, PartKey={1}, Message={2}", t.Amount, t.PartitionKey, transresult.Message);
          Core.Queues.QueueManager.AccountTransactionsLoggQueue.AddMessage(msg);

          Trace.WriteLine(msg.AsString, "Information");
          queue.DeleteMessage(msg);

        }
        else if (transresult.TransactionStatus == TransactionStatus.Completed)
        {
          Trace.WriteLine(String.Format("Amount:{0}, Message={1}, Time={2}, Error={3}", t.Amount, t.Message, t.Timestamp, transresult.Message), "Warning");
          queue.DeleteMessage(msg);
        }
      }
    }
  }
}
