﻿#region usings
using System;
using System.Linq;
using Core.DataContext;
using Core.TableStorage;
using System.Data.Services.Client;
using System.Net;
using System.Globalization;
using Core.Types;
using Core.Extensions;
using System.Xml.Serialization;
using System.IO;
using Microsoft.WindowsAzure.StorageClient;
using System.Collections.Specialized;
using Core.Models;
#endregion

namespace Core.Logic
{
  public class UserAccountStorageAction : Core.Logic.IUserAccountStorageAction
  {
    
    public TransactionResponse InsertOrRedrawMoneyWithTableStorage(UserAccountTransaction accountTransaction)
    {
      string[] keys = new string[2];
      TransactionResponse transactionResponse = 
        new TransactionResponse { TransactionStatus=TransactionStatus.UnKnown};
      
      if (accountTransaction.IsParitionkeyValid(out keys))
      {
        transactionResponse.Message = String.Format("Invalid partition keys {0}",accountTransaction.PartitionKey);
        transactionResponse.TransactionStatus = TransactionStatus.Failed;
        return transactionResponse;
      }

      using (var db = new UserAccountDataContext())
      {
        try
        {
          var transaction = this.TransactionExists(accountTransaction.PartitionKey, accountTransaction.RowKey);
          if (transaction != null)
          {
            transactionResponse.Message = 
              String.Format("Transaction exists tid:{1}, Message={0}, ", transaction.Message, transaction.RowKey);

            transactionResponse.TransactionStatus = TransactionStatus.Completed;
            this.LoggTransactionToTableStorage(
              new AccountTransactionLog(keys[0], keys[1]) { Message = transactionResponse.Message });

            return transactionResponse;
          }

          UserAccountBalance userAccount = null;

          try
          {
            userAccount = 
              db.UserAccountBalanceTable.Where(uid => uid.PartitionKey == keys[0].ToString() 
                                              && uid.RowKey == keys[1].ToString()).FirstOrDefault();
          }
          catch (DataServiceQueryException e)
          {
            HttpStatusCode s;
            if (TableStorageHelpers.EvaluateException(e, out s) && s == HttpStatusCode.NotFound)
            {
              transactionResponse.Message = String.Format("Unable to find user account {0}", accountTransaction.PartitionKey);
              transactionResponse.TransactionStatus = TransactionStatus.Failed;
              return transactionResponse;
            }
          }

          var balancetemp = userAccount.Balance + accountTransaction.Amount;
          if (balancetemp < 0)
          {
            transactionResponse.Message = String.Format("Not enough money, tried to redraw {0} but only {1} available", accountTransaction.Amount, userAccount.Balance);
            transactionResponse.TransactionStatus = TransactionStatus.Failed;
            this.LoggTransactionToTableStorage(new AccountTransactionLog(keys[0], keys[1]) { Message = transactionResponse.Message });
            return transactionResponse;
          }

          userAccount.Balance += accountTransaction.Amount;
          db.UpdateObject(userAccount);
          db.AddObject(UserAccountDataContext.TableNames.UserAccountTransactionTable.ToString(), accountTransaction);
          db.SaveChanges();
          transactionResponse.TransactionStatus = TransactionStatus.Succeded;

          return transactionResponse;
        }
        catch (DataServiceQueryException e)
        {
          HttpStatusCode s;
          if (TableStorageHelpers.EvaluateException(e, out s) && s == HttpStatusCode.NotFound)
          {
            // this would mean the entity was not found
            transactionResponse.TransactionStatus = TransactionStatus.Failed;
          }
          else
            transactionResponse.Message = GetExceptionMessage(e);
        }
      }
      return transactionResponse;
    }
   
    public void CreateUserAccount(UserAccountBalance userAccountTransaction)
    {
      using (var db = new UserAccountDataContext())
      {
        try
        {
          bool found = CheckIfAccountExists(userAccountTransaction, db);

          if (!found)
          {
            db.AddObject(UserAccountDataContext.TableNames.UserAccountBalanceTable.ToString(), userAccountTransaction);
            db.SaveChanges();
            //db.BeginSaveChanges(
            //                    async=>
            //                    {
            //                      db.EndSaveChanges(async);
            //                    },null);
          }
        }
        catch (DataServiceQueryException e)
        {throw new Exception(GetExceptionMessage(e));}
      }
    }
   
    public double GetAccountBalance(Guid userId, Guid accountId)
    {
      using (var db = new UserAccountDataContext())
      {
        //db.SendingRequest += new EventHandler<SendingRequestEventArgs>(db_SendingRequest);
        try
        {
          var balance = db.UserAccountBalanceTable.Where(
              uid => uid.PartitionKey == userId.ToString()
                   && uid.RowKey == accountId.ToString()).FirstOrDefault();
          
          return balance.Balance;
        }
        catch (DataServiceQueryException e)
        {
          HttpStatusCode s;
          if (TableStorageHelpers.EvaluateException(e, out s) && s == HttpStatusCode.NotFound)
          {
            // this would mean the entity was not found
            return -1 ;
          }
        }
        return -1;
      }
    }

    public void LoggTransactionToTableStorage(AccountTransactionLog msg)
    {
      try
      {
        var db = new UserAccountDataContext();
        db.AddObject(UserAccountDataContext.TableNames.AccountTransactionLoggTable.ToString(), msg);
        db.BeginSaveChanges(SaveChangesOptions.None, (asyncResult) =>
                            {
                              db.EndSaveChanges(asyncResult);
                            }, null);
      }
      catch (DataServiceQueryException e)
      {
        throw new Exception(GetExceptionMessage(e));
      }
    }

    public TransactionResponse ReliableHandleMoneyInQueueAsXml(UserAccountTransaction accountTransaction)
    {
      Core.Models.TransactionResponse response = 
        this.CheckIfTransactionExists(accountTransaction.PartitionKey, accountTransaction.RowKey);
      
      if(response.TransactionStatus== TransactionStatus.Proceed)
      {
        using (MemoryStream m = new MemoryStream())
        {
          XmlSerializer xs = new XmlSerializer(typeof(UserAccountTransaction));
          xs.Serialize(m, accountTransaction);
          try
          {
            Core.Queues.QueueManager.AccountTransactionsQueue.AddMessage(new CloudQueueMessage(m.ToArray()));
            response.TransactionStatus = TransactionStatus.Succeded;
          }
          catch(StorageClientException)
          {
            response.TransactionStatus = TransactionStatus.Failed;
            response.Message = String.Format("Unable to insert message in the account transaction queueu userId|AccountId={0}, messageId={1}", accountTransaction.PartitionKey, accountTransaction.RowKey);
          }
        }
      }
      return response;
    }
    
    public TransactionResponse ReliableHandleMoneyInQueue(UserAccountTransaction accountTransaction)
    {
      TransactionResponse response = this.CheckIfTransactionExists(accountTransaction.PartitionKey, accountTransaction.RowKey);
      if (response.TransactionStatus == TransactionStatus.Proceed)
      {
        //userid|accountid is partkey
        //userid|accountid|transactionid|amount is message
        string msg = string.Format("{0}|{1}|{2}", accountTransaction.PartitionKey, accountTransaction.RowKey, accountTransaction.Amount);
        try
        {
          Core.Queues.QueueManager.AccountTransactionsQueue.AddMessage(new CloudQueueMessage(msg));
          response.TransactionStatus = TransactionStatus.Succeded;
        }
        catch(StorageClientException)
        {
          response.TransactionStatus = TransactionStatus.Failed;
          response.Message = String.Format("Unable to insert message in the account transaction queueu userId|AccountId={0}, messageId={1}", accountTransaction.PartitionKey, accountTransaction.RowKey);
        }
      }
      return response;
    }

    public void DeleteUserAccount(Guid userId, Guid accountId)
    {
      using (var db = new UserAccountDataContext())
      {
        try
        {
          var accounts = db.UserAccountTransactionTable.Where(
                  uac => uac.PartitionKey == String.Format("{0}|{1}",userId,accountId));
          accounts.ForEach(a => db.DeleteObject(a));
          
          //remove all useraccounts from queues
        }
        catch (DataServiceQueryException e)
        { throw new Exception(GetExceptionMessage(e)); }
      }
    }

    #region private methods
    private TransactionResponse CheckIfTransactionExists(string userIdAccountId, string transId)
    {
      TransactionResponse transactionResponse = new TransactionResponse();

      var transaction = this.TransactionExists(userIdAccountId, transId);
      if (transaction != null)
      {
        transactionResponse.Message = String.Format("messageId:{0}, Message={1}, ", transaction.RowKey, transaction.Message);
        transactionResponse.TransactionStatus = TransactionStatus.Completed;
      }
      else
        transactionResponse.TransactionStatus = TransactionStatus.Proceed;

      return transactionResponse;

    }
    
    private UserAccountTransaction TransactionExists(string userIdAccountId, string transId)
    {
      UserAccountTransaction userAccountTransaction = null;
      using (var db = new UserAccountDataContext())
      {
        try
        {
          userAccountTransaction = db.UserAccountTransactionTable.Where(
                  uac => uac.PartitionKey == userIdAccountId && uac.RowKey == transId).FirstOrDefault();
          userAccountTransaction.Message = "Transaction Exists";
        }
        catch (DataServiceQueryException e)
        {
          HttpStatusCode s;
          if (TableStorageHelpers.EvaluateException(e, out s) && s == HttpStatusCode.NotFound)
          {
            // this would mean the entity was not found
            userAccountTransaction = null;
          }
        }
      }
      return userAccountTransaction;
    }
    
    private static string GetExceptionMessage(Exception exception)
    {
      HttpStatusCode statusCode;
      StorageExtendedErrorInformation extendedErrorInfo;
      if (TableStorageHelpers.EvaluateException(exception, out statusCode, out extendedErrorInfo))
      {
        if (extendedErrorInfo != null)
        {
          return string.Format(CultureInfo.InvariantCulture, "{0} {1}", extendedErrorInfo.ErrorCode ?? "", extendedErrorInfo.ErrorMessage ?? "");
        }
      }

      DataServiceClientException dse = exception.InnerException as DataServiceClientException;
      if (dse != null)
        return dse.Message;
      else
        return exception.Message;
    }
    
    private bool CheckIfAccountExists(UserAccountBalance userAccountTransaction, UserAccountDataContext db)
    {
      //ensure that same account does not exist
      bool found = false;
      try
      {
        var result = db.UserAccountBalanceTable.Where(
          x => x.PartitionKey == userAccountTransaction.PartitionKey && x.RowKey == userAccountTransaction.RowKey).FirstOrDefault();
        if (result != null)
          found = true;
      }
      catch (DataServiceQueryException e)
      {
        HttpStatusCode s;
        if (TableStorageHelpers.EvaluateException(e, out s) && s == HttpStatusCode.NotFound)
        {
          // this would mean the entity was not found
          found = false;
        }
      }
      return found;
    }
    
    void db_SendingRequest(object sender, SendingRequestEventArgs e)
    {
      //var a=e.Request.Headers[HttpRequestHeader.IfMatch];
      //a = e.Request.Headers[HttpRequestHeader.IfModifiedSince];
      //a=e.Request.Headers[HttpRequestHeader.IfNoneMatch];
    }
    #endregion
  }

  #region junk
  public static class TableStorageHelpers
  {

    #region Error handling helpers

    /// <summary>
    /// Checks whether the exception is or contains a DataServiceClientException and extracts the 
    /// returned http status code and extended error information.
    /// </summary>
    /// <param name="exception">The exception from which to extract information</param>
    /// <param name="status">The Http status code for the exception</param>
    /// <param name="extendedErrorInfo">Extended error information including storage service specific
    /// error code and error message</param>
    /// <returns>True if the exception is or contains a DataServiceClientException.</returns>
    public static bool EvaluateException(
        Exception exception,
        out HttpStatusCode status,
        out StorageExtendedErrorInformation extendedErrorInfo
        )
    {
      return EvaluateExceptionImpl(exception, out status, out extendedErrorInfo, true);
    }

    /// <summary>
    /// Checks whether the exception is or contains a DataServiceClientException and extracts the 
    /// returned http status code.
    /// </summary>
    /// <param name="exception">The exception from which to extract information</param>
    /// <param name="status">The Http status code for the exception</param>
    /// <returns>True if the exception is or contains a DataServiceClientException.</returns>
    public static bool EvaluateException(
        Exception exception,
        out HttpStatusCode status
        )
    {
      StorageExtendedErrorInformation extendedErrorInfo;
      return EvaluateExceptionImpl(exception, out status, out extendedErrorInfo, false);
    }

    private static bool EvaluateExceptionImpl(
        Exception e,
        out HttpStatusCode status,
        out StorageExtendedErrorInformation extendedErrorInfo,
        bool getExtendedErrors
        )
    {
      status = HttpStatusCode.Unused;
      extendedErrorInfo = null;
      while (e.InnerException != null)
      {
        e = e.InnerException;

        DataServiceClientException dsce = e as DataServiceClientException;
        if (dsce != null)
        {
          status = (HttpStatusCode)dsce.StatusCode;
          return true;
        }
      }
      return false;
    }

    /// <summary>
    /// Checks whether the exception is either a DataServiceClientException, a DataServiceQueryException or a 
    /// DataServiceRequestException.
    /// </summary>
    public static bool IsTableStorageException(Exception exception)
    {
      return ((exception is DataServiceClientException) || (exception is DataServiceQueryException) || (exception is DataServiceRequestException));
    }

    /// <summary>
    /// Only certain classes of errors should be retried. This method evaluates an exception 
    /// and returns whether this class of exception can be retried.
    /// </summary>
    /// <param name="e">The exception to analyze.</param>
    /// <param name="statusCode">The HttpStatusCode retrieved from the exception.</param>
    internal static bool CanBeRetried(InvalidOperationException e, out HttpStatusCode statusCode)
    {
      HttpStatusCode status;

      statusCode = HttpStatusCode.Unused;
      if (EvaluateException(e, out status))
      {
        statusCode = status;
        if (status == HttpStatusCode.RequestTimeout ||
          // server error codes above 500
            status == HttpStatusCode.ServiceUnavailable ||
            status == HttpStatusCode.InternalServerError ||
            status == HttpStatusCode.BadGateway ||
            status == HttpStatusCode.GatewayTimeout)
        {
          return true;
        }
      }
      return false;
    }

    /// <summary>
    /// Overload that does not retrun the HttpStatusCode.
    /// </summary>
    internal static bool CanBeRetried(InvalidOperationException e)
    {
      HttpStatusCode ignored;
      return CanBeRetried(e, out ignored);
    }

    #endregion

    #region Methods for checking properties to be inserted into a table

    /// <summary>
    /// Checks whether the string can be inserted in a table storage table. Throws an exception if 
    /// this is not the case.
    /// </summary>
    /// <param name="propertyValue"></param>
    public static void CheckStringProperty(string propertyValue)
    {
      if (string.IsNullOrEmpty(propertyValue))
      {
        throw new ArgumentException("The string cannot be null or empty!");
      }
      if (propertyValue.Length > TableStorageConstants.MaxStringPropertySizeInChars)
      {
        throw new ArgumentException("The string cannot be longer than the maximum string property size.");
      }
    }

    /// <summary>
    /// Checks whether the string can be inserted into a table storage table.
    /// </summary>
    public static bool ValidateStringProperty(string propertyValue)
    {
      if (string.IsNullOrEmpty(propertyValue))
      {
        return false;
      }
      if (propertyValue.Length > TableStorageConstants.MaxStringPropertySizeInChars)
      {
        return false;
      }
      return true;
    }

    #endregion

  }
  /// <summary>
  /// Class representing some important table storage constants.
  /// </summary>
  public static class TableStorageConstants
  {
    /// <summary>
    /// The maximum size of strings per property/column is 64 kB (that is 32k characters.)
    /// Note: This constant is smaller for the development storage table service.
    /// </summary>
    public static readonly int MaxStringPropertySizeInBytes = 64 * 1024;


    /// <summary>
    /// One character in the standard UTF-16 character presentation is 2 bytes.
    /// Note: This constant is smaller for the development storage table service.
    /// </summary>
    public static readonly int MaxStringPropertySizeInChars = MaxStringPropertySizeInBytes / 2;

    /// <summary>
    /// We want to prevent users from the pitfall of mixing up Utc and local time.
    /// Because of this we add some time to the minimum supported datetime.
    /// As a result, there will be no error condition from the server even 
    /// if a user converts the minimum supported date time to a local time and 
    /// stores this in a DateTime field.
    /// The local development storage support the SQL range of dates which is narrower than the
    /// one for the table storage service and so we use that value here. 
    /// </summary>
    public static readonly DateTime MinSupportedDateTime = DateTime.FromFileTime(0).ToUniversalTime().AddYears(200);

    //You can use this if you are programming against the real table storage service only but then your
    //code will not work against the local development table storage.
    //public static readonly DateTime MinSupportedDateTime = DateTime.FromFileTime(0).ToUniversalTime().AddDays(7);

    /// <summary>
    /// Internal constant for querying tables.
    /// </summary>
    internal const string TablesName = "Tables";

    /// <summary>
    /// Internal constant for querying tables.
    /// </summary>
    internal const string TablesQuery = "/" + TablesName;
  }
  [Serializable]
  public class StorageExtendedErrorInformation
  {
    public string ErrorCode { get; internal set; }
    public string ErrorMessage { get; internal set; }
    public NameValueCollection AdditionalDetails { get; internal set; }
  }
#endregion
}
